/*
Copyright (C) 2011 Observatoire de Paris-Meudon

Ce programme est un logiciel libre ; vous pouvez le redistribuer et/ou le modifier conformment aux dispositions de la Licence Publique Gnrale GNU, telle que publie par la Free Software Foundation ; version 2 de la licence, ou encore ( votre choix) toute version ultrieure.

Ce programme est distribu dans l'espoir qu'il sera utile, mais SANS AUCUNE GARANTIE ; sans mme la garantie implicite de COMMERCIALISATION ou D'ADAPTATION A UN OBJET PARTICULIER. Pour plus de dtail, voir la Licence Publique Gnrale GNU .

Vous devez avoir reu un exemplaire de la Licence Publique Gnrale GNU en mme temps que ce programme ; si ce n'est pas le cas, crivez  la Free Software Foundation Inc., 675 Mass Ave, Cambridge, MA 02139, Etats-Unis.
*/

package ecliptique;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.GridLayout;
import java.awt.Image;
import java.awt.RenderingHints;
import java.awt.geom.Arc2D;
import java.awt.geom.Ellipse2D;
import java.awt.geom.GeneralPath;
import java.awt.geom.Line2D;
import java.awt.geom.Point2D;
import java.util.Hashtable;
import javax.swing.BorderFactory;
import javax.swing.JApplet;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JSlider;
import javax.swing.SwingConstants;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;

public class AppletEcliptique extends JApplet implements ChangeListener {
    
    static final int minJour = 1;
    int maxJour = 31; // dpend du mois
    static final int minMois = 1;
    static final int maxMois = 12;
    static final int minAnnee = 2000;
    static final int maxAnnee = 2020;
    static final int jourInitial = 1;
    static final int moisInitial = 1;
    static final int anneeInitiale = 2000;
    static final Color couleurEquateur = new Color(150, 255, 200, 200);
    static final Color couleurTexteEquateur = new Color(100, 200, 150);
    static final Color couleurEcliptique = new Color(255, 220, 150, 200);
    static final Color couleurTexteEcliptique = new Color(220, 150, 80);
    static final Color couleurSoleil = new Color(255, 255, 0);
    static final Color couleurLambda = new Color(150, 100, 50);
    static final Color couleurBeta = new Color(150, 50, 150);
    static final Color couleurAlpha = new Color(255, 50, 50, 230);
    static final Color couleurDelta = new Color(50, 50, 255);
    static final Color couleurEpsilon = new Color(50, 130, 50);
    static final Color couleurArcs = Color.lightGray;
    static final int precisionCourbes = 100;
    int imgdimx;
    int imgdimy;
    protected ImageCanvas ic;
    protected Image croquis = null;
    protected Graphics2D g2d = null;
    protected JSlider sliderJour, sliderMois, sliderAnnee;
    protected double lambda, beta, alpha, delta, epsilon;
    protected int jour, mois, annee;
    protected int cx, cy; // position du centre de l'image en pixels
    protected int rayon; // rayon du grand cercle en pixels
    protected JLabel labelJour, labelMois, labelAnnee;
    protected JLabel labelInfos;
    protected Font gras;
    protected Matrice rotObservateur; // matrice de rotation par rapport  l'observateur
    protected Matrice rotEcliptique; // matrice de rotation de l'cliptique par rapport  l'quateur
    protected Matrice rotObsEcl; // rotObservateur * rotEcliptique
    protected Vecteur O, G, D, EA, P; // points en coordonnes sphriques
    protected String titreLambda, titreBeta, titreAlpha, titreDelta, titreEpsilon, titreGamma, titreEA;
    
    
    public void init() {
        imgdimx = Math.min(getWidth() - 200, getHeight() - 50);
        if (imgdimx <= 0)
            imgdimx = 500;
        imgdimy = imgdimx;
        getContentPane().setLayout(new BorderLayout());
        getContentPane().setBackground(Color.white);
        
        creerCroquis();
        ic = new ImageCanvas(imgdimx, imgdimy);
        getContentPane().add(ic, BorderLayout.WEST);
        labelInfos = new JLabel();
        getContentPane().add(labelInfos, BorderLayout.SOUTH);
        
        final JPanel controlPane = new JPanel(new BorderLayout());
        controlPane.setBorder(BorderFactory.createEmptyBorder(0, 10, 0, 0));
        final JPanel labelPane = new JPanel(new GridLayout(4, 1));
        labelPane.setPreferredSize(new Dimension(120, imgdimy));
        labelPane.setBackground(Color.white);
        controlPane.add(labelPane, BorderLayout.WEST);
        final JPanel sliderPane = new JPanel(new GridLayout(4, 1));
        sliderPane.setBackground(Color.white);
        controlPane.add(sliderPane, BorderLayout.CENTER);
        controlPane.setBackground(Color.white);
        getContentPane().add(controlPane, BorderLayout.CENTER);
        
        if (gras.canDisplay('\u03BB'))
            titreLambda = "\u03BB";
        else
            titreLambda = "lambda";
        
        if (gras.canDisplay('\u03B2'))
            titreBeta = "\u03B2";
        else
            titreBeta = "beta";
        
        if (gras.canDisplay('\u03B1'))
            titreAlpha = "\u03B1";
        else
            titreAlpha = "alpha";
        
        if (gras.canDisplay('\u03B4'))
            titreDelta = "\u03B4";
        else
            titreDelta = "delta";
        
        if (gras.canDisplay('\u03B5'))
            titreEpsilon = "\u03B5";
        else
            titreEpsilon = "eps";
        
        if (gras.canDisplay('\u03B3'))
            titreGamma = "\u03B3";
        else
            titreGamma = "Y";
        
        jour = jourInitial;
        labelJour = new JLabel("Jour = " + jour);
        labelJour.setFont(gras);
        labelPane.add(labelJour);
        sliderJour = new JSlider(minJour, maxJour, jour);
        sliderJour.setSnapToTicks(true);
        sliderJour.addChangeListener(this);
        sliderJour.setMinorTickSpacing(1);
        sliderJour.setPaintTicks(true);
        Hashtable graduations = new Hashtable();
        graduations.put(new Integer(0), new JLabel("0"));
        graduations.put(new Integer(maxJour), new JLabel(""+maxJour));
        sliderJour.setLabelTable(graduations);
        sliderJour.setPaintLabels(true);
        sliderJour.setBackground(Color.white);
        sliderPane.add(sliderJour);
        
        mois = moisInitial;
        labelMois = new JLabel("Mois = " + mois);
        labelMois.setFont(gras);
        labelPane.add(labelMois);
        sliderMois = new JSlider(minMois, maxMois, mois);
        sliderMois.setSnapToTicks(true);
        sliderMois.addChangeListener(this);
        sliderMois.setMinorTickSpacing(1);
        sliderMois.setPaintTicks(true);
        graduations = new Hashtable();
        graduations.put(new Integer(minMois), new JLabel(""+minMois));
        graduations.put(new Integer(maxMois), new JLabel(""+maxMois));
        sliderMois.setLabelTable(graduations);
        sliderMois.setPaintLabels(true);
        sliderMois.setBackground(Color.white);
        sliderPane.add(sliderMois);
        
        annee = anneeInitiale;
        labelAnnee = new JLabel("Anne = " + annee);
        labelAnnee.setFont(gras);
        labelPane.add(labelAnnee);
        sliderAnnee = new JSlider(minAnnee, maxAnnee, annee);
        sliderAnnee.setSnapToTicks(true);
        sliderAnnee.addChangeListener(this);
        sliderAnnee.setMajorTickSpacing(10);
        sliderAnnee.setMinorTickSpacing(1);
        sliderAnnee.setPaintTicks(true);
        graduations = new Hashtable();
        graduations.put(new Integer(minAnnee), new JLabel(""+minAnnee));
        graduations.put(new Integer(2010), new JLabel("2010"));
        graduations.put(new Integer(maxAnnee), new JLabel(""+maxAnnee));
        sliderAnnee.setLabelTable(graduations);
        sliderAnnee.setPaintLabels(true);
        sliderAnnee.setBackground(Color.white);
        sliderPane.add(sliderAnnee);
        
        rotObservateur = new Matrice(3, 3);
        rotEcliptique = new Matrice(3, 3);
        O = new Vecteur(0, 0, 0);
        G = new Vecteur(1, Math.PI / 2, -Math.PI / 2);
        D = new Vecteur(1, Math.PI / 2, Math.PI / 2);
        EA = new Vecteur(1, Math.PI / 2, Math.PI);
        P = new Vecteur(1, 0, 0);
        
        majCroquis();
        
        validate();
        if (ic != null)
            ic.requestFocus();
    }
    
    public void stateChanged(ChangeEvent e) {
        Object adj = e.getSource();
        if (adj == sliderJour) {
            int nJour = sliderJour.getValue();
            if (nJour != jour) {
                jour = nJour;
                labelJour.setText("Jour = " + jour);
                majCroquis();
            }
        } else if (adj == sliderMois) {
            int nMois = sliderMois.getValue();
            if (nMois != mois) {
                mois = nMois;
                labelMois.setText("Mois = " + mois);
                maxJour = joursParMois(mois, annee);
                sliderJour.setMaximum(maxJour);
                Hashtable graduations = new Hashtable();
                graduations.put(new Integer(0), new JLabel("0"));
                graduations.put(new Integer(maxJour), new JLabel(""+maxJour));
                sliderJour.setLabelTable(graduations);
                majCroquis();
            }
        } else if (adj == sliderAnnee) {
            int nAnnee = sliderAnnee.getValue();
            if (nAnnee != annee) {
                annee = nAnnee;
                labelAnnee.setText("Anne = " + annee);
                maxJour = joursParMois(mois, annee);
                sliderJour.setMaximum(maxJour);
                Hashtable graduations = new Hashtable();
                graduations.put(new Integer(0), new JLabel("0"));
                graduations.put(new Integer(maxJour), new JLabel(""+maxJour));
                sliderJour.setLabelTable(graduations);
                majCroquis();
            }
        }
    }
    
    protected void creerCroquis() {
        croquis = createImage(imgdimx, imgdimy);
        g2d = (Graphics2D)croquis.getGraphics();
        gras = new Font(getFont().getName(), Font.BOLD, 12);
        g2d.setFont(gras);
        g2d.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);
        g2d.setRenderingHint(RenderingHints.KEY_TEXT_ANTIALIASING, RenderingHints.VALUE_TEXT_ANTIALIAS_ON);
        g2d.setRenderingHint(RenderingHints.KEY_STROKE_CONTROL, RenderingHints.VALUE_STROKE_PURE);
    }
    
    protected void majCroquis() {
        // calcul du temps sidral
        // Jour Julien le 1/1/2000  0 h UTC : 2451544.5
        int jours = jour - 1; // jours couls depuis le 1/1/2000
        for (int i=2000; i<annee; i++)
            jours += joursParAn(i);
        for (int i=1; i<12; i++) {
            if (mois > i)
                jours += joursParMois(i, annee);
        }
        
        double T = (2451544.5 + jours - 2451545) / 36525.; // en siecles juliens depuis le 1/1/2000  12h (pour le jour donn et l'heure 0:00)
        // obliquit moyenne de l'cliptique, en degrs
        double secondes = 21.448 - T*(46.8150 + T*(0.00059 - T*(0.001813)));
        double e0 = 23.0 + (26.0 + (secondes/60.0))/60.0;
        
        double omega = Math.toRadians(125.04 - 1934.136 * T);
        // inclinaison de la Terre, en radians
        epsilon = Math.toRadians(e0 + 0.00256 * Math.cos(omega));
        
        // longitude cliptique moyenne du soleil, en degrs
        double Lm = 280.46646 + T * (36000.76983 + 0.0003032 * T);
        // anomalie moyenne du soleil, en radians
        double M1 = Math.toRadians(357.52911 + T * (35999.05029 - 0.0001537 * T));
        // quation du centre, en degrs
        double C = Math.sin(M1) * (1.914602 - T * (0.004817 + 0.000014 * T)) + Math.sin(2*M1) * (0.019993 - 0.000101 * T) + Math.sin(3*M1) * 0.000289;
        // longitude vraie, en degrs
        double lv = (Lm + C) % 360;
        // longitude apparente, en radians
        lambda = Math.toRadians(lv - 0.00569 - 0.00478 * Math.sin(omega));
        beta = 0; // approximation, si on nglige les effets de la Lune et des autres plantes
        
        final double aobs = 0.2; // angle d'observation
        rotObservateur.set(0, 0, Math.cos(aobs));
        rotObservateur.set(0, 1, 0);
        rotObservateur.set(0, 2, -Math.sin(aobs));
        rotObservateur.set(1, 0, 0);
        rotObservateur.set(1, 1, 1);
        rotObservateur.set(1, 2, 0);
        rotObservateur.set(2, 0, Math.sin(aobs));
        rotObservateur.set(2, 1, 0);
        rotObservateur.set(2, 2, Math.cos(aobs));
        
        rotEcliptique.set(0, 0, 1);
        rotEcliptique.set(0, 1, 0);
        rotEcliptique.set(0, 2, 0);
        rotEcliptique.set(1, 0, 0);
        rotEcliptique.set(1, 1, Math.cos(epsilon));
        rotEcliptique.set(1, 2, Math.sin(epsilon));
        rotEcliptique.set(2, 0, 0);
        rotEcliptique.set(2, 1, -Math.sin(epsilon));
        rotEcliptique.set(2, 2, Math.cos(epsilon));
        
        rotObsEcl = rotObservateur.produit(rotEcliptique);
        
        g2d.setColor(Color.white);
        g2d.fillRect(0, 0, imgdimx, imgdimy);
        cx = imgdimx / 2;
        cy = imgdimy / 2;
        rayon = Math.min(cx - 20, cy - 20); // rayon de la sphre en pixels
        
        g2d.setColor(new Color(230, 230, 255));
        g2d.fillOval(cx - rayon, cy - rayon, rayon * 2, rayon * 2);
        
        g2d.setColor(Color.darkGray);
        g2d.drawOval(cx - rayon, cy - rayon, rayon * 2, rayon * 2);
        
        final Vecteur M = new Vecteur(1, Math.PI / 2 - beta, lambda); // position du Soleil dans le repre de l'cliptique
        if (M.get(2) < -Math.PI)
            M.set(2, M.get(2) + 2 * Math.PI); // pour avoir la 3me coordonne entre -pi et pi
        final Vecteur Mcart = M.spheriquesVersCartesiennes();
        final Vecteur McartEq = rotEcliptique.produit(Mcart); // rotation cliptique -> quateur
        final Vecteur MsphEq = McartEq.cartesiennesVersSpheriques(); // coordonnes sphriques de M dans le repre de l'quateur
        alpha = MsphEq.get(2);
        if (alpha < 0)
            alpha += 2*Math.PI;
        delta = Math.PI/2 - MsphEq.get(1);
        final Point2D ptM = conversion(M, true);
        final Point2D ptO = new Point2D.Double(cx, cy);
        final Vecteur B = new Vecteur(1, Math.PI / 2, lambda); // repre cliptique
        if (B.get(2) < -Math.PI)
            B.set(2, B.get(2) + 2 * Math.PI);
        final Point2D ptB = conversion(B, true);
        final Point2D ptG = conversion(G, false);
        final Vecteur K = new Vecteur(1, epsilon, -Math.PI / 2);
        //final Point2D ptK = conversion(K, false);
        final Vecteur A = new Vecteur(1, Math.PI / 2, MsphEq.get(2)); // repre de l'quateur
        final Vecteur pointVernal = new Vecteur(1, Math.PI / 2, 0); // repre cliptique ET quateur !
        Point2D pt;
        
        // calcul cercle de l'cliptique (derrire)
        final GeneralPath cheminEcl1 = new GeneralPath();
        pt = conversion(new Vecteur(1, Math.PI / 2, 0), true);
        cheminEcl1.moveTo((float)pt.getX(), (float)pt.getY());
        for (int i=1; i<=precisionCourbes; i++) {
            pt = conversion(new Vecteur(1, Math.PI / 2, - (i / (double)precisionCourbes) * Math.PI), true);
            cheminEcl1.lineTo((float)pt.getX(), (float)pt.getY());
        }
        cheminEcl1.closePath();
        
        // arc PMA
        dessinArcPMA(P, MsphEq, A, false);
        
        // arc PMB
        dessinArcPMB(P, M, B, false);
        
        dessinSoleil(M, ptM, MsphEq, false);
        
        dessinAngleBeta(B, M, false);
        
        dessinAngleDelta(A, MsphEq, false);
        
        // affichage cercle de l'cliptique (derrire)
        g2d.setColor(couleurEcliptique);
        g2d.fill(cheminEcl1);
        
        dessinAngleLambda(pointVernal, B, false);
        
        // angle alpha (partie derrire l'horizon)
        if (MsphEq.get(1) > 1E-5 && MsphEq.get(1) < Math.PI - 1E-5) // toile au ple: alpha indtermin
            dessinAngleAlpha(pointVernal, A, false);
        
        dessinEquateur(pointVernal);
        
        // calcul cercle de l'cliptique (devant)
        final GeneralPath cheminEcl2 = new GeneralPath();
        pt = conversion(new Vecteur(1, Math.PI / 2, 0), true);
        cheminEcl2.moveTo((float)pt.getX(), (float)pt.getY());
        for (int i=1; i<=precisionCourbes/2; i++) {
            pt = conversion(new Vecteur(1, Math.PI / 2, (i / (double)(precisionCourbes/2)) * Math.PI), true);
            cheminEcl2.lineTo((float)pt.getX(), (float)pt.getY());
        }
        cheminEcl2.closePath();
        // affichage cercle de l'cliptique (devant)
        g2d.setColor(couleurEcliptique);
        g2d.fill(cheminEcl2);
        g2d.setColor(couleurTexteEcliptique);
        if (epsilon < Math.PI/2) {
            pt = conversion(pointVernal, false);
            g2d.rotate(-epsilon, pt.getX(), pt.getY());
            g2d.drawString("Ecliptique", (float)(pt.getX() - rayon * 0.75), (float)pt.getY() + 10);
            g2d.rotate(epsilon, pt.getX(), pt.getY());
        } else {
            pt = conversion(EA, false);
            g2d.rotate(epsilon, pt.getX(), pt.getY());
            g2d.drawString("Ecliptique", (float)(pt.getX() - rayon * 0.75), (float)pt.getY() - 5);
            g2d.rotate(-epsilon, pt.getX(), pt.getY());
        }
        
        dessinArcPMA(P, MsphEq, A, true);
        
        dessinArcPMB(P, M, B, true);
        
        dessinSoleil(M, ptM, MsphEq, true);
        
        dessinAngleLambda(pointVernal, B, true);
        
        dessinAngleBeta(B, M, true);
        
        dessinAngleEpsilon();
        
        // angle alpha (partie devant l'horizon)
        if (MsphEq.get(1) > 1E-5 && MsphEq.get(1) < Math.PI - 1E-5) // toile au ple: alpha indtermin
            dessinAngleAlpha(pointVernal, A, true);
        
        dessinAngleDelta(A, MsphEq, true);
        
        // points
        g2d.setColor(Color.black);
        
        // affichage des points fixes
        afficherPoint(O, false, true, "O");
        //afficherPoint(G, false, true, "G");
        //afficherPoint(D, false, true, "D");
        //afficherPoint(EA, false, true, "quinoxe d'automne");
        afficherPoint(P, false, true, "P");
        
        // affichage des points en mouvement
        afficherPoint(K, false, true, "K");
        
        //afficherPoint(B, true, true, "B");
        //afficherPoint(A, false, true, "A");
        //afficherPoint(M, true, true, "M");
        afficherPoint(pointVernal, true, true, titreGamma);
        
        ic.updateImage(croquis);
        
        
        String infos = titreLambda + "=" + affichageAngleDMS(lambda, false, true);
        if (MsphEq.get(1) > 1E-5 && MsphEq.get(1) < Math.PI - 1E-5)
            infos += "   " + titreAlpha + "=" + affichageAngleHMS(alpha);
        infos += "   " + titreDelta + "=" + affichageAngleDMS(Math.PI/2 - MsphEq.get(1), false, false);
        labelInfos.setText(infos);
    }
    
    protected void dessinEquateur(final Vecteur pointVernal) {
        final GeneralPath cheminEquateur = new GeneralPath();
        Point2D pt = conversion(new Vecteur(1, Math.PI / 2, 0), false);
        cheminEquateur.moveTo((float)pt.getX(), (float)pt.getY());
        for (int i=1; i<precisionCourbes/2; i++) {
            pt = conversion(new Vecteur(1, Math.PI / 2, (i / (double)(precisionCourbes/2)) * (2 * Math.PI)), false);
            cheminEquateur.lineTo((float)pt.getX(), (float)pt.getY());
        }
        cheminEquateur.closePath();
        g2d.setColor(couleurEquateur);
        g2d.fill(cheminEquateur);
        g2d.setColor(couleurTexteEquateur);
        pt = conversion(pointVernal, false);
        g2d.rotate(-0.1, pt.getX(), pt.getY());
        g2d.drawString("Equateur", (float)(pt.getX() + rayon / 4), (float)pt.getY() + 20);
        g2d.rotate(0.1, pt.getX(), pt.getY());
    }
    
    protected void dessinSoleil(final Vecteur M, final Point2D ptM, final Vecteur MsphEq, final boolean devant) {
        final boolean Mdevant = (Math.cos(M.get(2)) > 0);
        if (Mdevant == devant) {
            final Ellipse2D cercle = new Ellipse2D.Double(ptM.getX() - 5, ptM.getY() - 5, 10, 10);
            g2d.setColor(Color.black);
            g2d.draw(cercle);
            g2d.setColor(couleurSoleil);
            g2d.fill(cercle);
        }
    }
    
    protected void dessinArcPMA(final Vecteur P, final Vecteur MsphEq, final Vecteur A, final boolean devant) {
        final boolean Adevant = (Math.cos(A.get(2)) > 0);
        if (Adevant == devant) {
            g2d.setColor(couleurArcs);
            final Vecteur Pbis = new Vecteur(1, 0, MsphEq.get(2)); // P avec le mme phi que SsphEq
            dessinArc(Pbis, MsphEq.get(1) < A.get(1) ? MsphEq : A, true, false, false);
        }
    }
    
    protected void dessinArcPMB(final Vecteur P, final Vecteur M, final Vecteur B, final boolean devant) {
        final boolean Bdevant = (Math.cos(B.get(2)) > 0);
        if (Bdevant == devant) {
            g2d.setColor(couleurArcs);
            final Vecteur PEq = new Vecteur(1, 0, M.get(2)); // P dans le repre quatorial, avec le mme phi que S
            dessinArc(PEq, M.get(1) < B.get(1) ? M : B, true, true, false);
        }
    }
    
    protected void dessinAngleLambda(final Vecteur pointVernal, final Vecteur B, final boolean devant) {
        g2d.setColor(couleurLambda);
        final boolean Bdevant = (Math.cos(B.get(2)) > 0);
        final boolean Bderriere = (Math.cos(B.get(2)) <= 0);
        // dbut angle
        if (devant && (Bderriere || (lambda > Math.PI && Bdevant))) {
            // trait gamma-D (devant)
            dessinArc(pointVernal, D, true, true, false);
        }
        // milieu
        if (!devant && lambda > Math.PI && Bdevant) {
            // trait D-EA-G (derrire)
            dessinArc(new Vecteur(1, Math.PI/2, Math.PI/2), new Vecteur(1, Math.PI/2, 3*Math.PI/2), true, true, false);
        }
        // fin avec la flche
        if (devant && lambda < Math.PI && Bdevant) {
            // flche gamma-B (devant)
            dessinArc(pointVernal, B, true, true, true);
        } else if (devant && lambda > Math.PI && Bdevant) {
            // flche G-B (devant)
            dessinArc(new Vecteur(1, Math.PI/2, -Math.PI/2), B, true, true, true);
        } else if (!devant && Bderriere) {
            // flche D-B (derrire)
            dessinArc(new Vecteur(1, Math.PI/2, Math.PI/2), B, true, true, true);
        }
        // titre
        if (devant) {
            double moyennePhi;
            if (B.get(2) - pointVernal.get(2) < 0)
                moyennePhi = pointVernal.get(2) + (B.get(2) + 2*Math.PI - pointVernal.get(2))/2;
            else
                moyennePhi = pointVernal.get(2) + (B.get(2) - pointVernal.get(2))/2;
            Point2D pt = conversion(new Vecteur(1, Math.PI/2, moyennePhi), true);
            if (epsilon < Math.PI/2)
                g2d.drawString(titreLambda, (float)(pt.getX() - 10), (float)(pt.getY() - 5));
            else
                g2d.drawString(titreLambda, (float)(pt.getX()), (float)(pt.getY() - 5));
        }
    }
    
    protected void dessinAngleBeta(final Vecteur B, final Vecteur M, final boolean devant) {
        g2d.setColor(couleurBeta);
        final boolean Bdevant = (Math.cos(B.get(2)) > 0);
        if (devant == Bdevant) {
            dessinArc(B, M, true, true, true);
            if (Math.abs(beta) > 0.1) {
                double r;
                if ((epsilon >= 0 && beta >= 0) || (epsilon < 0 && beta < 0))
                    r = 1.1;
                else
                    r = 0.9;
                Point2D pt = conversion(new Vecteur(r, Math.PI/2 - (Math.PI/2 - M.get(1))/2, B.get(2)), true);
                g2d.drawString(titreBeta, (float)(pt.getX() - g2d.getFontMetrics().stringWidth(titreBeta)/2), (float)(pt.getY() + 5));
            }
        }
    }
    
    protected void dessinAngleEpsilon() {
        g2d.setColor(couleurEpsilon);
        dessinArc(new Vecteur(1.01, D.get(1), D.get(2)), new Vecteur(1.01, Math.PI/2 - epsilon, Math.PI/2), true, false, true);
        Point2D pt = conversion(new Vecteur(1.05, Math.PI/2 - epsilon, Math.PI/2), false);
        g2d.drawString(titreEpsilon, (float)(pt.getX() - g2d.getFontMetrics().stringWidth(titreEpsilon)/2), (float)(pt.getY() + 5));
    }
    
    protected void dessinAngleAlpha(final Vecteur pointVernal, final Vecteur A, final boolean devant) {
        final boolean Adevant = (Math.cos(A.get(2)) > 0);
        final boolean Aderriere = (Math.cos(A.get(2)) <= 0);
        g2d.setColor(couleurAlpha);
        // dbut angle
        if (devant && (Aderriere || (alpha > Math.PI && Adevant))) {
            // trait gamma-D (devant)
            dessinArc(pointVernal, D, true, false, false);
        }
        // milieu
        if (!devant && alpha > Math.PI && Adevant) {
            // trait D-EA-G (derrire)
            dessinArc(new Vecteur(1, Math.PI/2, Math.PI/2), new Vecteur(1, Math.PI/2, 3*Math.PI/2), true, false, false);
        }
        // fin avec la flche
        if (devant && alpha < Math.PI && Adevant) {
            // flche gamma-A (devant)
            dessinArc(pointVernal, A, true, false, true);
        } else if (devant && alpha > Math.PI && Adevant) {
            // flche G-A (devant)
            dessinArc(new Vecteur(1, Math.PI/2, -Math.PI/2), A, true, false, true);
        } else if (!devant && Aderriere) {
            // flche D-A (derrire)
            dessinArc(new Vecteur(1, Math.PI/2, Math.PI/2), A, true, false, true);
        }
        
        if (devant) {
            Point2D pt = conversion(new Vecteur(1, Math.PI/2, 0.75*alpha), false);
            g2d.drawString(titreAlpha, (float)(pt.getX() - g2d.getFontMetrics().stringWidth(titreAlpha)/2), (float)(pt.getY() - 5));
        }
    }
    
    protected void dessinAngleDelta(final Vecteur A, final Vecteur MsphEq, final boolean devant) {
        final boolean Adevant = (Math.cos(A.get(2)) > 0);
        if (Adevant == devant) {
            g2d.setColor(couleurDelta);
            dessinArc(A, MsphEq, true, false, true);
            if (Math.abs(MsphEq.get(1) - Math.PI/2) > 0.1) {
                final Point2D pt = conversion(new Vecteur(1, Math.PI/2 - (Math.PI/2 - MsphEq.get(1))/2, A.get(2)), false);
                double x;
                if (beta >= 0)
                    x = pt.getX() - g2d.getFontMetrics().stringWidth(titreDelta) - 5;
                else
                    x = pt.getX() + 5;
                g2d.drawString(titreDelta, (float)(x), (float)(pt.getY() + 5));
            }
        }
    }
    
    // dessin d'un arc entre 2 points dans l'un des 2 systmes de coordonnes sphriques, en suivant l'une des coordonnes
    protected void dessinArc(final Vecteur v1, final Vecteur v2, final boolean sensPhiPositif, final boolean coordEcl, final boolean fleche) {
        final int nb = precisionCourbes/4;
        final double dr = (v2.get(0) - v1.get(0)) / nb;
        final double dtheta = (v2.get(1) - v1.get(1)) / nb;
        double phi1 = v1.get(2);
        double phi2 = v2.get(2);
        if (v1.get(1) == 0)
            phi1 = phi2;
        else if (v2.get(1) == 0)
            phi2 = phi1;
        if (phi1 < 0)
            phi1 += 2 * Math.PI;
        if (phi2 < 0)
            phi2 += 2 * Math.PI;
        if (sensPhiPositif && phi1 > phi2)
            phi2 += 2 * Math.PI;
        else if (!sensPhiPositif && phi1 < phi2)
            phi1 += 2 * Math.PI;
        final double dphi = (phi2 - phi1) / nb; // ne marche pas toujours (par ex. avec theta = 0)
        final GeneralPath chemin = new GeneralPath();
        Point2D pt = conversion(v1, coordEcl);
        Point2D vpt = null;
        chemin.moveTo((float)pt.getX(), (float)pt.getY());
        for (int i=1; i<=nb; i++) {
            if (i == nb)
                vpt = pt;
            pt = conversion(new Vecteur(v1.get(0) + i * dr, v1.get(1) + i * dtheta, phi1 + i * dphi), coordEcl);
            chemin.lineTo((float)pt.getX(), (float)pt.getY());
        }
        g2d.draw(chemin);
        if (fleche && Math.abs(v2.get(0) - v1.get(0)) + Math.abs(v2.get(1) - v1.get(1)) + Math.abs(phi2 - phi1) > 0)
            dessinFleche(vpt, pt);
    }
    
    // dessin des 2 traits de la fin de la flche  partir des 2 derniers points d'un arc
    protected void dessinFleche(final Point2D pt1, final Point2D pt2) {
        final double l = 6; // longueur des traits en pixels
        final double da = Math.PI / 8; // angle des traits avec la fin de l'arc
        final double dx = pt2.getX() - pt1.getX();
        final double dy = pt2.getY() - pt1.getY();
        double angle;
        if (dx == 0) {
            if (dy > 0)
                angle = Math.PI / 2;
            else
                angle = -Math.PI / 2;
        } else {
            angle = Math.atan(dy / dx);
            if (dx < 0)
                angle = angle + Math.PI;
        }
        Point2D ptFleche = new Point2D.Double(pt2.getX() - l * Math.cos(angle + da), pt2.getY() - l * Math.sin(angle + da));
        g2d.draw(new Line2D.Double(pt2.getX(), pt2.getY(), ptFleche.getX(), ptFleche.getY()));
        ptFleche = new Point2D.Double(pt2.getX() - l * Math.cos(angle - da), pt2.getY() - l * Math.sin(angle - da));
        g2d.draw(new Line2D.Double(pt2.getX(), pt2.getY(), ptFleche.getX(), ptFleche.getY()));
    }
    
    // passage des coordonnes sphriques aux coordonnes cran
    protected Point2D conversion(final Vecteur vsph, final boolean rotEq) {
        // passage aux coordonnes cartsiennes
        Vecteur v = vsph.spheriquesVersCartesiennes();
        
        // rotation optionelle de l'quateur et rotation permettant de voir l'horizon de haut
        if (rotEq)
            v = rotObsEcl.produit(v);
        else
            v = rotObservateur.produit(v);
        
        return(projectionEcran(v));
    }
    
    // ajout d'un point en coordonnes sphriques (r, theta, phi)
    protected void afficherPoint(final Vecteur vsph, final boolean rotEq, final boolean croix, final String titre) {
        final Point2D p = conversion(vsph, rotEq);
        
        // affichage d'une croix ou d'un point
        if (croix) {
            g2d.draw(new Line2D.Double(p.getX() - 2, p.getY(), p.getX() + 2, p.getY()));
            g2d.draw(new Line2D.Double(p.getX(), p.getY() - 2, p.getX(), p.getY() + 2));
        } else
            g2d.draw(new Line2D.Double(p.getX(), p.getY(), p.getX(), p.getY()));
        
        // affichage du titre
        if (titre != null) {
            final int l = g2d.getFontMetrics().stringWidth(titre);
            g2d.drawString(titre, (float)p.getX() - l/2, (float)p.getY() - 9);
        }
    }
    
    // projection selon l'axe x  partir des coordonnes cartsiennes, puis conversion vers l'cran
    protected Point2D projectionEcran(final Vecteur v) {
        return(new Point2D.Double(cx + v.get(1) * rayon, cy - v.get(2) * rayon));
    }
    
    // affichage en heures minutes d'un angle en radians > 0
    protected static String affichageAngleHMS(final double a) {
        double h = a*12/Math.PI;
        int ih = (int)Math.floor(h);
        double m = (h - ih)*60;
        int im = (int)Math.round(m);
        if (im >= 60) {
            im -= 60;
            ih += 1;
        }
        return(ih + "h " + im + "m");
    }
    
    // affichage en degrs minutes d'un angle en radians
    protected static String affichageAngleDMS(final double a, final boolean estOuest, final boolean angle360) {
        double d = Math.toDegrees(Math.abs(a));
        int id = (int)Math.floor(d);
        double m = (d - id)*60;
        int im = (int)Math.round(m);
        if (im >= 60) {
            im -= 60;
            id += 1;
        }
        String signe;
        if (angle360 || a == 0)
            signe = "";
        else if (a > 0)
            signe = "+";
        else
            signe = "-";
        if (estOuest)
            return(id + " " + im + "'" + (a >= 0 ? " E" : " O"));
        else
            return(signe + id + " " + im + "'");
    }
    
    protected static int joursParMois(final int mois, final int an) {
        switch (mois) {
            case 1: return(31);
            case 2: {
                if (an % 4 == 0 && (an % 100 != 0 || an % 400 == 0))
                    return(29);
                else
                    return(28);
            }
            case 3: return(31);
            case 4: return(30);
            case 5: return(31);
            case 6: return(30);
            case 7: return(31);
            case 8: return(31);
            case 9: return(30);
            case 10: return(31);
            case 11: return(30);
            case 12: return(31);
        }
        return(0);
    }
    
    protected static int joursParAn(final int an) {
        if (an % 4 == 0 && (an % 100 != 0 || an % 400 == 0))
            return(366);
        else
            return(365);
    }
}
