/*
Copyright (C) 2011 Observatoire de Paris-Meudon

Ce programme est un logiciel libre ; vous pouvez le redistribuer et/ou le modifier conformment aux dispositions de la Licence Publique Gnrale GNU, telle que publie par la Free Software Foundation ; version 2 de la licence, ou encore ( votre choix) toute version ultrieure.

Ce programme est distribu dans l'espoir qu'il sera utile, mais SANS AUCUNE GARANTIE ; sans mme la garantie implicite de COMMERCIALISATION ou D'ADAPTATION A UN OBJET PARTICULIER. Pour plus de dtail, voir la Licence Publique Gnrale GNU .

Vous devez avoir reu un exemplaire de la Licence Publique Gnrale GNU en mme temps que ce programme ; si ce n'est pas le cas, crivez  la Free Software Foundation Inc., 675 Mass Ave, Cambridge, MA 02139, Etats-Unis.
*/

package ecliptique;

public class Matrice {
    double[][] m;
    
    public Matrice(final int lignes, final int colonnes) {
        m = new double[colonnes][lignes];
    }
    
    public Vecteur produit(final Vecteur v) {
        final int nbc = m.length;
        final int nbl = m[0].length;
        if (nbc != v.getDimensions()) {
            System.err.println("produit impossible d'une matrice avec un vecteur (" + nbc + " != " + v.getDimensions() + ")");
            return(null);
        }
        final Vecteur res = new Vecteur(nbl);
        double d;
        for (int j=0; j<nbl; j++) {
            d = 0;
            for (int i=0; i<nbc; i++)
                d += v.get(i) * m[i][j];
            res.set(j, d);
        }
        return(res);
    }
    
    public Matrice produit(final Matrice mat) {
        final int nbc = m.length;
        final int nbl = m[0].length;
        if (nbc != mat.getLignes()) {
            System.err.println("produit impossible d'une matrice avec une matrice (" + nbc + " != " + mat.getLignes() + ")");
            return(null);
        }
        final Matrice res = new Matrice(nbl, nbc);
        double d;
        for (int jr=0; jr<nbc; jr++) {
            for (int ir=0; ir<nbl; ir++) {
                d = 0;
                for (int i=0; i<nbc; i++)
                    d += m[i][jr] * mat.get(ir, i);
                res.set(ir, jr, d);
            }
        }
        return(res);
    }
    
    public int getColonnes() {
        return(m.length);
    }
    
    public int getLignes() {
        return(m[0].length);
    }
    
    public double get(final int colonne, final int ligne) {
        return(m[colonne][ligne]);
    }
    
    public void set(final int colonne, final int ligne, final double valeur) {
        m[colonne][ligne] = valeur;
    }
}
